@@ -6,21 +6,24 @@ use num_traits::One;
 use crate::{
     args::CallArgs,
     color::Color,
-    common::QuoteKind,
+    common::{Brackets, ListSeparator, QuoteKind},
     error::SassResult,
     parse::Parser,
     unit::Unit,
     value::{Number, Value},
 };
 
-fn inner_hsl(name: &'static str, mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+fn inner_hsl(name: &'static str, mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     if args.is_empty() {
         return Err(("Missing argument $channels.", args.span()).into());
     }
 
-    if args.len() == 1 {
+    let len = args.len();
+
+    if len == 1 {
         let mut channels = match args.get_err(0, "channels")? {
             Value::List(v, ..) => v,
+            v if v.is_special_function() => vec![v],
             _ => return Err(("Missing argument $channels.", args.span()).into()),
         };
 
@@ -35,6 +38,24 @@ fn inner_hsl(name: &'static str, mut args: CallArgs, parser: &mut Parser<'_>) ->
                 .into());
         }
 
+        if channels.iter().any(Value::is_special_function) {
+            let channel_sep = if channels.len() < 3 {
+                ListSeparator::Space
+            } else {
+                ListSeparator::Comma
+            };
+
+            return Ok(Value::String(
+                format!(
+                    "{}({})",
+                    name,
+                    Value::List(channels, channel_sep, Brackets::None)
+                        .to_css_string(args.span(), false)?
+                ),
+                QuoteKind::None,
+            ));
+        }
+
         let lightness = match channels.pop() {
             Some(Value::Dimension(Some(n), ..)) => n / Number::from(100),
             Some(Value::Dimension(None, ..)) => todo!(),
@@ -81,26 +102,42 @@ fn inner_hsl(name: &'static str, mut args: CallArgs, parser: &mut Parser<'_>) ->
             Number::one(),
         ))))
     } else {
-        let hue = match args.get_err(0, "hue")? {
+        let hue = args.get_err(0, "hue")?;
+        let saturation = args.get_err(1, "saturation")?;
+        let lightness = args.get_err(2, "lightness")?;
+        let alpha = args.default_arg(
+            3,
+            "alpha",
+            Value::Dimension(Some(Number::one()), Unit::None, true),
+        )?;
+
+        if [&hue, &saturation, &lightness, &alpha]
+            .iter()
+            .copied()
+            .any(Value::is_special_function)
+        {
+            return Ok(Value::String(
+                format!(
+                    "{}({})",
+                    name,
+                    Value::List(
+                        if len == 4 {
+                            vec![hue, saturation, lightness, alpha]
+                        } else {
+                            vec![hue, saturation, lightness]
+                        },
+                        ListSeparator::Comma,
+                        Brackets::None
+                    )
+                    .to_css_string(args.span(), false)?
+                ),
+                QuoteKind::None,
+            ));
+        }
+
+        let hue = match hue {
             Value::Dimension(Some(n), ..) => n,
             Value::Dimension(None, ..) => todo!(),
-            v if v.is_special_function() => {
-                let saturation = args.get_err(1, "saturation")?;
-                let lightness = args.get_err(2, "lightness")?;
-                let mut string = format!(
-                    "{}({}, {}, {}",
-                    name,
-                    v.to_css_string(args.span())?,
-                    saturation.to_css_string(args.span())?,
-                    lightness.to_css_string(args.span())?
-                );
-                if !args.is_empty() {
-                    string.push_str(", ");
-                    string.push_str(&args.get_err(3, "alpha")?.to_css_string(args.span())?);
-                }
-                string.push(')');
-                return Ok(Value::String(string, QuoteKind::None));
-            }
             v => {
                 return Err((
                     format!("$hue: {} is not a number.", v.inspect(args.span())?),
@@ -109,70 +146,35 @@ fn inner_hsl(name: &'static str, mut args: CallArgs, parser: &mut Parser<'_>) ->
                     .into())
             }
         };
-        let saturation = match args.get_err(1, "saturation")? {
+        let saturation = match saturation {
             Value::Dimension(Some(n), ..) => n / Number::from(100),
             Value::Dimension(None, ..) => todo!(),
-            v if v.is_special_function() => {
-                let lightness = args.get_err(2, "lightness")?;
-                let mut string = format!(
-                    "{}({}, {}, {}",
-                    name,
-                    hue,
-                    v.to_css_string(args.span())?,
-                    lightness.to_css_string(args.span())?
-                );
-                if !args.is_empty() {
-                    string.push_str(", ");
-                    string.push_str(&args.get_err(3, "alpha")?.to_css_string(args.span())?);
-                }
-                string.push(')');
-                return Ok(Value::String(string, QuoteKind::None));
-            }
             v => {
                 return Err((
                     format!(
                         "$saturation: {} is not a number.",
-                        v.to_css_string(args.span())?
+                        v.to_css_string(args.span(), parser.options.is_compressed())?
                     ),
                     args.span(),
                 )
                     .into())
             }
         };
-        let lightness = match args.get_err(2, "lightness")? {
+        let lightness = match lightness {
             Value::Dimension(Some(n), ..) => n / Number::from(100),
             Value::Dimension(None, ..) => todo!(),
-            v if v.is_special_function() => {
-                let mut string = format!(
-                    "{}({}, {}, {}",
-                    name,
-                    hue,
-                    saturation,
-                    v.to_css_string(args.span())?
-                );
-                if !args.is_empty() {
-                    string.push_str(", ");
-                    string.push_str(&args.get_err(3, "alpha")?.to_css_string(args.span())?);
-                }
-                string.push(')');
-                return Ok(Value::String(string, QuoteKind::None));
-            }
             v => {
                 return Err((
                     format!(
                         "$lightness: {} is not a number.",
-                        v.to_css_string(args.span())?
+                        v.to_css_string(args.span(), parser.options.is_compressed())?
                     ),
                     args.span(),
                 )
                     .into())
             }
         };
-        let alpha = match args.default_arg(
-            3,
-            "alpha",
-            Value::Dimension(Some(Number::one()), Unit::None, true),
-        )? {
+        let alpha = match alpha {
             Value::Dimension(Some(n), Unit::None, _) => n,
             Value::Dimension(Some(n), Unit::Percent, _) => n / Number::from(100),
             Value::Dimension(None, ..) => todo!(),
@@ -180,25 +182,12 @@ fn inner_hsl(name: &'static str, mut args: CallArgs, parser: &mut Parser<'_>) ->
                 return Err((
                     format!(
                         "$alpha: Expected {} to have no units or \"%\".",
-                        v.to_css_string(args.span())?
+                        v.to_css_string(args.span(), parser.options.is_compressed())?
                     ),
                     args.span(),
                 )
                     .into())
             }
-            v if v.is_special_function() => {
-                return Ok(Value::String(
-                    format!(
-                        "{}({}, {}, {}, {})",
-                        name,
-                        hue,
-                        saturation,
-                        lightness,
-                        v.to_css_string(args.span())?
-                    ),
-                    QuoteKind::None,
-                ));
-            }
             v => {
                 return Err((
                     format!("$alpha: {} is not a number.", v.inspect(args.span())?),
@@ -213,15 +202,15 @@ fn inner_hsl(name: &'static str, mut args: CallArgs, parser: &mut Parser<'_>) ->
     }
 }
 
-pub(crate) fn hsl(args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn hsl(args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     inner_hsl("hsl", args, parser)
 }
 
-pub(crate) fn hsla(args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn hsla(args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     inner_hsl("hsla", args, parser)
 }
 
-pub(crate) fn hue(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn hue(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(1)?;
     match args.get_err(0, "color")? {
         Value::Color(c) => Ok(Value::Dimension(Some(c.hue()), Unit::Deg, true)),
@@ -233,7 +222,7 @@ pub(crate) fn hue(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Val
     }
 }
 
-pub(crate) fn saturation(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn saturation(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(1)?;
     match args.get_err(0, "color")? {
         Value::Color(c) => Ok(Value::Dimension(Some(c.saturation()), Unit::Percent, true)),
@@ -245,7 +234,7 @@ pub(crate) fn saturation(mut args: CallArgs, parser: &mut Parser<'_>) -> SassRes
     }
 }
 
-pub(crate) fn lightness(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn lightness(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(1)?;
     match args.get_err(0, "color")? {
         Value::Color(c) => Ok(Value::Dimension(Some(c.lightness()), Unit::Percent, true)),
@@ -257,7 +246,7 @@ pub(crate) fn lightness(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResu
     }
 }
 
-pub(crate) fn adjust_hue(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn adjust_hue(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(2)?;
     let color = match args.get_err(0, "color")? {
         Value::Color(c) => c,
@@ -276,7 +265,7 @@ pub(crate) fn adjust_hue(mut args: CallArgs, parser: &mut Parser<'_>) -> SassRes
             return Err((
                 format!(
                     "$degrees: {} is not a number.",
-                    v.to_css_string(args.span())?
+                    v.to_css_string(args.span(), parser.options.is_compressed())?
                 ),
                 args.span(),
             )
@@ -286,7 +275,7 @@ pub(crate) fn adjust_hue(mut args: CallArgs, parser: &mut Parser<'_>) -> SassRes
     Ok(Value::Color(Box::new(color.adjust_hue(degrees))))
 }
 
-fn lighten(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+fn lighten(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(2)?;
     let color = match args.get_err(0, "color")? {
         Value::Color(c) => c,
@@ -305,7 +294,7 @@ fn lighten(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
             return Err((
                 format!(
                     "$amount: {} is not a number.",
-                    v.to_css_string(args.span())?
+                    v.to_css_string(args.span(), false)?
                 ),
                 args.span(),
             )
@@ -315,7 +304,7 @@ fn lighten(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
     Ok(Value::Color(Box::new(color.lighten(amount))))
 }
 
-fn darken(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+fn darken(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(2)?;
     let color = match args.get_err(0, "color")? {
         Value::Color(c) => c,
@@ -334,7 +323,7 @@ fn darken(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
             return Err((
                 format!(
                     "$amount: {} is not a number.",
-                    v.to_css_string(args.span())?
+                    v.to_css_string(args.span(), false)?
                 ),
                 args.span(),
             )
@@ -344,13 +333,14 @@ fn darken(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
     Ok(Value::Color(Box::new(color.darken(amount))))
 }
 
-fn saturate(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+fn saturate(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(2)?;
     if args.len() == 1 {
         return Ok(Value::String(
             format!(
                 "saturate({})",
-                args.get_err(0, "amount")?.to_css_string(args.span())?
+                args.get_err(0, "amount")?
+                    .to_css_string(args.span(), false)?
             ),
             QuoteKind::None,
         ));
@@ -363,7 +353,7 @@ fn saturate(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
             return Err((
                 format!(
                     "$amount: {} is not a number.",
-                    v.to_css_string(args.span())?
+                    v.to_css_string(args.span(), false)?
                 ),
                 args.span(),
             )
@@ -374,7 +364,7 @@ fn saturate(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
         Value::Color(c) => c,
         Value::Dimension(Some(n), u, _) => {
             return Ok(Value::String(
-                format!("saturate({}{})", n, u),
+                format!("saturate({}{})", n.inspect(), u),
                 QuoteKind::None,
             ))
         }
@@ -389,7 +379,7 @@ fn saturate(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
     Ok(Value::Color(Box::new(color.saturate(amount))))
 }
 
-fn desaturate(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+fn desaturate(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(2)?;
     let color = match args.get_err(0, "color")? {
         Value::Color(c) => c,
@@ -408,7 +398,7 @@ fn desaturate(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value>
             return Err((
                 format!(
                     "$amount: {} is not a number.",
-                    v.to_css_string(args.span())?
+                    v.to_css_string(args.span(), parser.options.is_compressed())?
                 ),
                 args.span(),
             )
@@ -418,13 +408,13 @@ fn desaturate(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value>
     Ok(Value::Color(Box::new(color.desaturate(amount))))
 }
 
-pub(crate) fn grayscale(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn grayscale(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(1)?;
     let color = match args.get_err(0, "color")? {
         Value::Color(c) => c,
         Value::Dimension(Some(n), u, _) => {
             return Ok(Value::String(
-                format!("grayscale({}{})", n, u),
+                format!("grayscale({}{})", n.inspect(), u),
                 QuoteKind::None,
             ))
         }
@@ -439,7 +429,7 @@ pub(crate) fn grayscale(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResu
     Ok(Value::Color(Box::new(color.desaturate(Number::one()))))
 }
 
-pub(crate) fn complement(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn complement(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(1)?;
     let color = match args.get_err(0, "color")? {
         Value::Color(c) => c,
@@ -454,7 +444,7 @@ pub(crate) fn complement(mut args: CallArgs, parser: &mut Parser<'_>) -> SassRes
     Ok(Value::Color(Box::new(color.complement())))
 }
 
-pub(crate) fn invert(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<Value> {
+pub(crate) fn invert(mut args: CallArgs, parser: &mut Parser) -> SassResult<Value> {
     args.max_args(2)?;
     let weight = match args.get(1, "weight") {
         Some(Err(e)) => return Err(e),
@@ -471,7 +461,7 @@ pub(crate) fn invert(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<
             return Err((
                 format!(
                     "$weight: {} is not a number.",
-                    v.to_css_string(args.span())?
+                    v.to_css_string(args.span(), parser.options.is_compressed())?
                 ),
                 args.span(),
             )
@@ -491,7 +481,7 @@ pub(crate) fn invert(mut args: CallArgs, parser: &mut Parser<'_>) -> SassResult<
                     .into());
             }
             Ok(Value::String(
-                format!("invert({}{})", n, u),
+                format!("invert({}{})", n.inspect(), u),
                 QuoteKind::None,
             ))
         }
