@@ -6,7 +6,6 @@ use std::{
 
 use clap::{arg_enum, App, AppSettings, Arg};
 
-#[cfg(not(feature = "wasm"))]
 use grass::{from_path, from_string, Options, OutputStyle};
 
 // TODO remove this
@@ -26,10 +25,6 @@ arg_enum! {
     }
 }
 
-#[cfg(feature = "wasm")]
-fn main() {}
-
-#[cfg(not(feature = "wasm"))]
 #[cfg_attr(feature = "profiling", inline(never))]
 fn main() -> std::io::Result<()> {
     let matches = App::new("grass")
