@@ -21,27 +21,92 @@ struct ToplevelUnknownAtRule {
     name: String,
     params: String,
     body: Vec<Stmt>,
+    has_body: bool,
+    is_group_end: bool,
+    inside_rule: bool,
+}
+
+#[derive(Debug, Clone)]
+struct BlockEntryUnknownAtRule {
+    name: String,
+    params: String,
 }
 
 #[derive(Debug, Clone)]
 enum Toplevel {
-    RuleSet(Selector, Vec<BlockEntry>),
+    RuleSet {
+        selector: Selector,
+        body: Vec<BlockEntry>,
+        is_group_end: bool,
+    },
     MultilineComment(String),
     UnknownAtRule(Box<ToplevelUnknownAtRule>),
     Keyframes(Box<Keyframes>),
     KeyframesRuleSet(Vec<KeyframesSelector>, Vec<BlockEntry>),
-    Media { query: String, body: Vec<Stmt> },
-    Supports { params: String, body: Vec<Stmt> },
-    Newline,
+    Media {
+        query: String,
+        body: Vec<Stmt>,
+        inside_rule: bool,
+        is_group_end: bool,
+    },
+    Supports {
+        params: String,
+        body: Vec<Stmt>,
+        inside_rule: bool,
+        is_group_end: bool,
+    },
     // todo: do we actually need a toplevel style variant?
     Style(Style),
     Import(String),
+    Empty,
+}
+
+impl Toplevel {
+    pub fn is_invisible(&self) -> bool {
+        match self {
+            Toplevel::RuleSet { selector, body, .. } => selector.is_empty() || body.is_empty(),
+            Toplevel::Media { body, .. } => body.is_empty(),
+            Toplevel::Empty => true,
+            _ => false,
+        }
+    }
+
+    pub fn is_group_end(&self) -> bool {
+        match self {
+            Toplevel::RuleSet { is_group_end, .. } => *is_group_end,
+            Toplevel::UnknownAtRule(t) => t.is_group_end && t.inside_rule,
+            Toplevel::Media {
+                inside_rule,
+                is_group_end,
+                ..
+            }
+            | Toplevel::Supports {
+                inside_rule,
+                is_group_end,
+                ..
+            } => *inside_rule && *is_group_end,
+            _ => false,
+        }
+    }
+}
+
+fn set_group_end(group: &mut [Toplevel]) {
+    match group.last_mut() {
+        Some(Toplevel::RuleSet { is_group_end, .. })
+        | Some(Toplevel::Supports { is_group_end, .. })
+        | Some(Toplevel::Media { is_group_end, .. }) => {
+            *is_group_end = true;
+        }
+        Some(Toplevel::UnknownAtRule(t)) => t.is_group_end = true,
+        _ => {}
+    }
 }
 
 #[derive(Debug, Clone)]
 enum BlockEntry {
     Style(Style),
     MultilineComment(String),
+    UnknownAtRule(BlockEntryUnknownAtRule),
 }
 
 impl BlockEntry {
@@ -49,13 +114,24 @@ impl BlockEntry {
         match self {
             BlockEntry::Style(s) => s.to_string(),
             BlockEntry::MultilineComment(s) => Ok(format!("/*{}*/", s)),
+            BlockEntry::UnknownAtRule(BlockEntryUnknownAtRule { name, params }) => {
+                Ok(if params.is_empty() {
+                    format!("@{};", name)
+                } else {
+                    format!("@{} {};", name, params)
+                })
+            }
         }
     }
 }
 
 impl Toplevel {
-    const fn new_rule(selector: Selector) -> Self {
-        Toplevel::RuleSet(selector, Vec::new())
+    const fn new_rule(selector: Selector, is_group_end: bool) -> Self {
+        Toplevel::RuleSet {
+            selector,
+            is_group_end,
+            body: Vec::new(),
+        }
     }
 
     fn new_keyframes_rule(selector: Vec<KeyframesSelector>) -> Self {
@@ -66,16 +142,28 @@ impl Toplevel {
         if s.value.is_null() {
             return;
         }
-        if let Toplevel::RuleSet(_, entries) | Toplevel::KeyframesRuleSet(_, entries) = self {
-            entries.push(BlockEntry::Style(s));
+
+        if let Toplevel::RuleSet { body, .. } | Toplevel::KeyframesRuleSet(_, body) = self {
+            body.push(BlockEntry::Style(s));
         } else {
             panic!();
         }
     }
 
     fn push_comment(&mut self, s: String) {
-        if let Toplevel::RuleSet(_, entries) | Toplevel::KeyframesRuleSet(_, entries) = self {
-            entries.push(BlockEntry::MultilineComment(s));
+        if let Toplevel::RuleSet { body, .. } | Toplevel::KeyframesRuleSet(_, body) = self {
+            body.push(BlockEntry::MultilineComment(s));
+        } else {
+            panic!();
+        }
+    }
+
+    fn push_unknown_at_rule(&mut self, at_rule: ToplevelUnknownAtRule) {
+        if let Toplevel::RuleSet { body, .. } = self {
+            body.push(BlockEntry::UnknownAtRule(BlockEntryUnknownAtRule {
+                name: at_rule.name,
+                params: at_rule.params,
+            }));
         } else {
             panic!();
         }
@@ -85,16 +173,16 @@ impl Toplevel {
 #[derive(Debug, Clone)]
 pub(crate) struct Css {
     blocks: Vec<Toplevel>,
-    in_at_rule: bool,
+    at_rule_context: AtRuleContext,
     allows_charset: bool,
     plain_imports: Vec<Toplevel>,
 }
 
 impl Css {
-    pub const fn new(in_at_rule: bool, allows_charset: bool) -> Self {
+    pub const fn new(at_rule_context: AtRuleContext, allows_charset: bool) -> Self {
         Css {
             blocks: Vec::new(),
-            in_at_rule,
+            at_rule_context,
             allows_charset,
             plain_imports: Vec::new(),
         }
@@ -102,23 +190,27 @@ impl Css {
 
     pub(crate) fn from_stmts(
         s: Vec<Stmt>,
-        in_at_rule: bool,
+        at_rule_context: AtRuleContext,
         allows_charset: bool,
     ) -> SassResult<Self> {
-        Css::new(in_at_rule, allows_charset).parse_stylesheet(s)
+        Css::new(at_rule_context, allows_charset).parse_stylesheet(s)
     }
 
     fn parse_stmt(&mut self, stmt: Stmt) -> SassResult<Vec<Toplevel>> {
         Ok(match stmt {
             Stmt::RuleSet { selector, body } => {
                 if body.is_empty() {
-                    return Ok(Vec::new());
+                    return Ok(vec![Toplevel::Empty]);
                 }
+
                 let selector = selector.into_selector().remove_placeholders();
+
                 if selector.is_empty() {
-                    return Ok(Vec::new());
+                    return Ok(vec![Toplevel::Empty]);
                 }
-                let mut vals = vec![Toplevel::new_rule(selector)];
+
+                let mut vals = vec![Toplevel::new_rule(selector, false)];
+
                 for rule in body {
                     match rule {
                         Stmt::RuleSet { .. } => vals.extend(self.parse_stmt(rule)?),
@@ -126,26 +218,54 @@ impl Css {
                         Stmt::Comment(s) => vals.first_mut().unwrap().push_comment(s),
                         Stmt::Media(m) => {
                             let MediaRule { query, body, .. } = *m;
-                            vals.push(Toplevel::Media { query, body });
+                            vals.push(Toplevel::Media {
+                                query,
+                                body,
+                                inside_rule: true,
+                                is_group_end: false,
+                            });
                         }
                         Stmt::Supports(s) => {
                             let SupportsRule { params, body } = *s;
-                            vals.push(Toplevel::Supports { params, body });
+                            vals.push(Toplevel::Supports {
+                                params,
+                                body,
+                                inside_rule: true,
+                                is_group_end: false,
+                            });
                         }
                         Stmt::UnknownAtRule(u) => {
                             let UnknownAtRule {
-                                params, body, name, ..
+                                params,
+                                body,
+                                name,
+                                has_body,
+                                ..
                             } = *u;
-                            vals.push(Toplevel::UnknownAtRule(Box::new(ToplevelUnknownAtRule {
+
+                            let at_rule = ToplevelUnknownAtRule {
                                 name,
                                 params,
                                 body,
-                            })));
+                                has_body,
+                                inside_rule: true,
+                                is_group_end: false,
+                            };
+
+                            if has_body {
+                                vals.push(Toplevel::UnknownAtRule(Box::new(at_rule)));
+                            } else {
+                                vals.first_mut().unwrap().push_unknown_at_rule(at_rule);
+                            }
                         }
                         Stmt::Return(..) => unreachable!(),
                         Stmt::AtRoot { body } => {
                             body.into_iter().try_for_each(|r| -> SassResult<()> {
-                                vals.append(&mut self.parse_stmt(r)?);
+                                let mut stmts = self.parse_stmt(r)?;
+
+                                set_group_end(&mut stmts);
+
+                                vals.append(&mut stmts);
                                 Ok(())
                             })?;
                         }
@@ -173,24 +293,47 @@ impl Css {
             Stmt::Style(s) => vec![Toplevel::Style(s)],
             Stmt::Media(m) => {
                 let MediaRule { query, body, .. } = *m;
-                vec![Toplevel::Media { query, body }]
+                vec![Toplevel::Media {
+                    query,
+                    body,
+                    inside_rule: false,
+                    is_group_end: false,
+                }]
             }
             Stmt::Supports(s) => {
                 let SupportsRule { params, body } = *s;
-                vec![Toplevel::Supports { params, body }]
+                vec![Toplevel::Supports {
+                    params,
+                    body,
+                    inside_rule: false,
+                    is_group_end: false,
+                }]
             }
             Stmt::UnknownAtRule(u) => {
                 let UnknownAtRule {
-                    params, body, name, ..
+                    params,
+                    body,
+                    name,
+                    has_body,
+                    ..
                 } = *u;
                 vec![Toplevel::UnknownAtRule(Box::new(ToplevelUnknownAtRule {
                     name,
                     params,
                     body,
+                    has_body,
+                    inside_rule: false,
+                    is_group_end: false,
                 }))]
             }
             Stmt::Return(..) => unreachable!("@return: {:?}", stmt),
-            Stmt::AtRoot { .. } => unreachable!("@at-root: {:?}", stmt),
+            Stmt::AtRoot { body } => body
+                .into_iter()
+                .map(|r| self.parse_stmt(r))
+                .collect::<SassResult<Vec<Vec<Toplevel>>>>()?
+                .into_iter()
+                .flatten()
+                .collect(),
             Stmt::Keyframes(k) => vec![Toplevel::Keyframes(k)],
             Stmt::KeyframesRuleSet(k) => {
                 let KeyframesRuleSet { body, selector } = *k;
@@ -211,20 +354,12 @@ impl Css {
     }
 
     fn parse_stylesheet(mut self, stmts: Vec<Stmt>) -> SassResult<Css> {
-        let mut is_first = true;
         for stmt in stmts {
-            let v = self.parse_stmt(stmt)?;
-            // this is how we print newlines between unrelated styles
-            // it could probably be refactored
-            if !v.is_empty() {
-                if let Some(Toplevel::MultilineComment(..)) = v.first() {
-                } else if is_first {
-                    is_first = false;
-                } else {
-                    self.blocks.push(Toplevel::Newline);
-                }
-                self.blocks.extend(v);
-            }
+            let mut v = self.parse_stmt(stmt)?;
+
+            set_group_end(&mut v);
+
+            self.blocks.extend(v);
         }
 
         // move plain imports to top of file
@@ -241,8 +376,15 @@ impl Css {
             OutputStyle::Compressed => {
                 CompressedFormatter::default().write_css(&mut buf, self, map)?;
             }
-            OutputStyle::Expanded => ExpandedFormatter::default().write_css(&mut buf, self, map)?,
+            OutputStyle::Expanded => {
+                ExpandedFormatter::default().write_css(&mut buf, self, map)?;
+
+                if !buf.is_empty() {
+                    writeln!(buf)?;
+                }
+            }
         }
+
         // TODO: check for this before writing
         let show_charset = allows_charset && buf.iter().any(|s| !s.is_ascii());
         let out = unsafe { String::from_utf8_unchecked(buf) };
@@ -265,11 +407,11 @@ trait Formatter {
 struct CompressedFormatter {}
 
 impl Formatter for CompressedFormatter {
-    fn write_css(&mut self, buf: &mut Vec<u8>, mut css: Css, map: &CodeMap) -> SassResult<()> {
-        for block in mem::take(&mut css.blocks) {
+    fn write_css(&mut self, buf: &mut Vec<u8>, css: Css, map: &CodeMap) -> SassResult<()> {
+        for block in css.blocks {
             match block {
-                Toplevel::RuleSet(selector, styles) => {
-                    if styles.is_empty() {
+                Toplevel::RuleSet { selector, body, .. } => {
+                    if body.is_empty() {
                         continue;
                     }
 
@@ -283,7 +425,7 @@ impl Formatter for CompressedFormatter {
                     }
 
                     write!(buf, "{{")?;
-                    self.write_block_entry(buf, &styles)?;
+                    self.write_block_entry(buf, &body)?;
                     write!(buf, "}}")?;
                 }
                 Toplevel::KeyframesRuleSet(selectors, styles) => {
@@ -293,30 +435,24 @@ impl Formatter for CompressedFormatter {
 
                     let mut selectors = selectors.iter();
                     if let Some(selector) = selectors.next() {
-                        match selector {
-                            KeyframesSelector::To => write!(buf, "to")?,
-                            KeyframesSelector::From => write!(buf, "from")?,
-                            KeyframesSelector::Percent(p) => write!(buf, "{}%", p)?,
-                        }
+                        write!(buf, "{}", selector)?;
                     }
                     for selector in selectors {
-                        match selector {
-                            KeyframesSelector::To => write!(buf, ",to")?,
-                            KeyframesSelector::From => write!(buf, ",from")?,
-                            KeyframesSelector::Percent(p) => write!(buf, ",{}%", p)?,
-                        }
+                        write!(buf, ",{}", selector)?;
                     }
 
                     write!(buf, "{{")?;
                     self.write_block_entry(buf, &styles)?;
                     write!(buf, "}}")?;
                 }
-                Toplevel::MultilineComment(..) => continue,
+                Toplevel::Empty | Toplevel::MultilineComment(..) => continue,
                 Toplevel::Import(s) => {
                     write!(buf, "@import {};", s)?;
                 }
                 Toplevel::UnknownAtRule(u) => {
-                    let ToplevelUnknownAtRule { params, name, body } = *u;
+                    let ToplevelUnknownAtRule {
+                        params, name, body, ..
+                    } = *u;
 
                     if params.is_empty() {
                         write!(buf, "@{}", name)?;
@@ -330,7 +466,7 @@ impl Formatter for CompressedFormatter {
                     }
 
                     write!(buf, "{{")?;
-                    let css = Css::from_stmts(body, true, css.allows_charset)?;
+                    let css = Css::from_stmts(body, AtRuleContext::Unknown, css.allows_charset)?;
                     self.write_css(buf, css, map)?;
                     write!(buf, "}}")?;
                 }
@@ -349,11 +485,11 @@ impl Formatter for CompressedFormatter {
                     }
 
                     write!(buf, "{{")?;
-                    let css = Css::from_stmts(body, true, css.allows_charset)?;
+                    let css = Css::from_stmts(body, AtRuleContext::Keyframes, css.allows_charset)?;
                     self.write_css(buf, css, map)?;
                     write!(buf, "}}")?;
                 }
-                Toplevel::Supports { params, body } => {
+                Toplevel::Supports { params, body, .. } => {
                     if params.is_empty() {
                         write!(buf, "@supports")?;
                     } else {
@@ -366,25 +502,24 @@ impl Formatter for CompressedFormatter {
                     }
 
                     write!(buf, "{{")?;
-                    let css = Css::from_stmts(body, true, css.allows_charset)?;
+                    let css = Css::from_stmts(body, AtRuleContext::Supports, css.allows_charset)?;
                     self.write_css(buf, css, map)?;
                     write!(buf, "}}")?;
                 }
-                Toplevel::Media { query, body } => {
+                Toplevel::Media { query, body, .. } => {
                     if body.is_empty() {
                         continue;
                     }
 
                     write!(buf, "@media {}{{", query)?;
-                    let css = Css::from_stmts(body, true, css.allows_charset)?;
+                    let css = Css::from_stmts(body, AtRuleContext::Media, css.allows_charset)?;
                     self.write_css(buf, css, map)?;
                     write!(buf, "}}")?;
                 }
                 Toplevel::Style(style) => {
-                    let value = style.value.node.to_css_string(style.value.span)?;
+                    let value = style.value.node.to_css_string(style.value.span, true)?;
                     write!(buf, "{}:{};", style.property, value)?;
                 }
-                Toplevel::Newline => {}
             }
         }
         Ok(())
@@ -410,25 +545,27 @@ impl CompressedFormatter {
     fn write_block_entry(&self, buf: &mut Vec<u8>, styles: &[BlockEntry]) -> SassResult<()> {
         let mut styles = styles.iter();
 
-        while let Some(style) = styles.next() {
+        for style in &mut styles {
             match style {
                 BlockEntry::Style(s) => {
-                    let value = s.value.node.to_css_string(s.value.span)?;
+                    let value = s.value.node.to_css_string(s.value.span, true)?;
                     write!(buf, "{}:{}", s.property, value)?;
                     break;
                 }
                 BlockEntry::MultilineComment(..) => continue,
+                b @ BlockEntry::UnknownAtRule(_) => write!(buf, "{}", b.to_string()?)?,
             }
         }
 
         for style in styles {
             match style {
                 BlockEntry::Style(s) => {
-                    let value = s.value.node.to_css_string(s.value.span)?;
+                    let value = s.value.node.to_css_string(s.value.span, true)?;
 
                     write!(buf, ";{}:{}", s.property, value)?;
                 }
                 BlockEntry::MultilineComment(..) => continue,
+                b @ BlockEntry::UnknownAtRule(_) => write!(buf, "{}", b.to_string()?)?,
             }
         }
         Ok(())
@@ -440,34 +577,60 @@ struct ExpandedFormatter {
     nesting: usize,
 }
 
+#[derive(Clone, Copy)]
+struct Previous {
+    is_group_end: bool,
+}
+
+/// What kind of @-rule are we currently inside
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub(crate) enum AtRuleContext {
+    Media,
+    Supports,
+    Keyframes,
+    Unknown,
+    None,
+}
+
 impl Formatter for ExpandedFormatter {
-    fn write_css(&mut self, buf: &mut Vec<u8>, mut css: Css, map: &CodeMap) -> SassResult<()> {
-        let mut has_written = false;
+    fn write_css(&mut self, buf: &mut Vec<u8>, css: Css, map: &CodeMap) -> SassResult<()> {
         let padding = "  ".repeat(self.nesting);
-        let mut should_emit_newline = false;
         self.nesting += 1;
-        for block in mem::take(&mut css.blocks) {
+
+        let mut prev: Option<Previous> = None;
+
+        for block in css.blocks {
+            if block.is_invisible() {
+                continue;
+            }
+
+            let is_group_end = block.is_group_end();
+
+            if let Some(prev) = prev {
+                writeln!(buf)?;
+
+                if (prev.is_group_end && css.at_rule_context == AtRuleContext::None)
+                    || css.at_rule_context == AtRuleContext::Supports
+                {
+                    writeln!(buf)?;
+                }
+            }
+
             match block {
-                Toplevel::RuleSet(selector, styles) => {
-                    if styles.is_empty() {
-                        continue;
-                    }
-                    has_written = true;
-                    if should_emit_newline && !css.in_at_rule {
-                        should_emit_newline = false;
-                        writeln!(buf)?;
-                    }
+                Toplevel::Empty => continue,
+                Toplevel::RuleSet { selector, body, .. } => {
                     writeln!(buf, "{}{} {{", padding, selector)?;
-                    for style in styles {
+
+                    for style in body {
                         writeln!(buf, "{}  {}", padding, style.to_string()?)?;
                     }
-                    writeln!(buf, "{}}}", padding)?;
+
+                    write!(buf, "{}}}", padding)?;
                 }
                 Toplevel::KeyframesRuleSet(selector, body) => {
                     if body.is_empty() {
                         continue;
                     }
-                    has_written = true;
 
                     writeln!(
                         buf,
@@ -482,22 +645,23 @@ impl Formatter for ExpandedFormatter {
                     for style in body {
                         writeln!(buf, "{}  {}", padding, style.to_string()?)?;
                     }
-                    writeln!(buf, "{}}}", padding)?;
+                    write!(buf, "{}}}", padding)?;
                 }
                 Toplevel::MultilineComment(s) => {
-                    has_written = true;
-                    writeln!(buf, "{}/*{}*/", padding, s)?;
+                    write!(buf, "{}/*{}*/", padding, s)?;
                 }
                 Toplevel::Import(s) => {
-                    has_written = true;
-                    writeln!(buf, "{}@import {};", padding, s)?;
+                    write!(buf, "{}@import {};", padding, s)?;
                 }
                 Toplevel::UnknownAtRule(u) => {
-                    let ToplevelUnknownAtRule { params, name, body } = *u;
-                    if should_emit_newline {
-                        should_emit_newline = false;
-                        writeln!(buf)?;
-                    }
+                    let ToplevelUnknownAtRule {
+                        params,
+                        name,
+                        body,
+                        has_body,
+                        inside_rule,
+                        ..
+                    } = *u;
 
                     if params.is_empty() {
                         write!(buf, "{}@{}", padding, name)?;
@@ -505,22 +669,34 @@ impl Formatter for ExpandedFormatter {
                         write!(buf, "{}@{} {}", padding, name, params)?;
                     }
 
-                    if body.is_empty() {
-                        writeln!(buf, ";")?;
+                    let css = Css::from_stmts(
+                        body,
+                        if inside_rule {
+                            AtRuleContext::Unknown
+                        } else {
+                            AtRuleContext::None
+                        },
+                        css.allows_charset,
+                    )?;
+
+                    if !has_body {
+                        write!(buf, ";")?;
+                        prev = Some(Previous { is_group_end });
+                        continue;
+                    }
+
+                    if css.blocks.iter().all(Toplevel::is_invisible) {
+                        write!(buf, " {{}}")?;
+                        prev = Some(Previous { is_group_end });
                         continue;
                     }
 
                     writeln!(buf, " {{")?;
-                    let css = Css::from_stmts(body, true, css.allows_charset)?;
                     self.write_css(buf, css, map)?;
-                    writeln!(buf, "{}}}", padding)?;
+                    write!(buf, "\n{}}}", padding)?;
                 }
                 Toplevel::Keyframes(k) => {
                     let Keyframes { rule, name, body } = *k;
-                    if should_emit_newline {
-                        should_emit_newline = false;
-                        writeln!(buf)?;
-                    }
 
                     write!(buf, "{}@{}", padding, rule)?;
 
@@ -529,21 +705,22 @@ impl Formatter for ExpandedFormatter {
                     }
 
                     if body.is_empty() {
-                        writeln!(buf, " {{}}")?;
+                        write!(buf, " {{}}")?;
+                        prev = Some(Previous { is_group_end });
                         continue;
                     }
 
                     writeln!(buf, " {{")?;
-                    let css = Css::from_stmts(body, true, css.allows_charset)?;
+                    let css = Css::from_stmts(body, AtRuleContext::Keyframes, css.allows_charset)?;
                     self.write_css(buf, css, map)?;
-                    writeln!(buf, "{}}}", padding)?;
+                    write!(buf, "\n{}}}", padding)?;
                 }
-                Toplevel::Supports { params, body } => {
-                    if should_emit_newline {
-                        should_emit_newline = false;
-                        writeln!(buf)?;
-                    }
-
+                Toplevel::Supports {
+                    params,
+                    body,
+                    inside_rule,
+                    ..
+                } => {
                     if params.is_empty() {
                         write!(buf, "{}@supports", padding)?;
                     } else {
@@ -551,37 +728,53 @@ impl Formatter for ExpandedFormatter {
                     }
 
                     if body.is_empty() {
-                        writeln!(buf, ";")?;
+                        write!(buf, ";")?;
+                        prev = Some(Previous { is_group_end });
                         continue;
                     }
 
                     writeln!(buf, " {{")?;
-                    let css = Css::from_stmts(body, true, css.allows_charset)?;
+                    let css = Css::from_stmts(
+                        body,
+                        if inside_rule {
+                            AtRuleContext::Supports
+                        } else {
+                            AtRuleContext::None
+                        },
+                        css.allows_charset,
+                    )?;
                     self.write_css(buf, css, map)?;
-                    writeln!(buf, "{}}}", padding)?;
+                    write!(buf, "\n{}}}", padding)?;
                 }
-                Toplevel::Media { query, body } => {
-                    if body.is_empty() {
-                        continue;
-                    }
-
+                Toplevel::Media {
+                    query,
+                    body,
+                    inside_rule,
+                    ..
+                } => {
                     writeln!(buf, "{}@media {} {{", padding, query)?;
-                    let css = Css::from_stmts(body, true, css.allows_charset)?;
+                    let css = Css::from_stmts(
+                        body,
+                        if inside_rule {
+                            AtRuleContext::Media
+                        } else {
+                            AtRuleContext::None
+                        },
+                        css.allows_charset,
+                    )?;
                     self.write_css(buf, css, map)?;
-                    writeln!(buf, "{}}}", padding)?;
+                    write!(buf, "\n{}}}", padding)?;
                 }
                 Toplevel::Style(s) => {
-                    writeln!(buf, "{}{}", padding, s.to_string()?)?;
-                }
-                Toplevel::Newline => {
-                    if has_written {
-                        should_emit_newline = true;
-                    }
-                    continue;
+                    write!(buf, "{}{}", padding, s.to_string()?)?;
                 }
             }
+
+            prev = Some(Previous { is_group_end });
         }
+
         self.nesting -= 1;
+
         Ok(())
     }
 }
