@@ -1,9 +1,3 @@
-use std::vec::IntoIter;
-
-use peekmore::PeekMoreIterator;
-
-use crate::Token;
-
 #[derive(Debug)]
 pub(crate) struct ParsedNumber {
     /// The full number excluding the decimal
@@ -45,15 +39,3 @@ impl ParsedNumber {
         }
     }
 }
-
-pub(crate) fn eat_whole_number(toks: &mut PeekMoreIterator<IntoIter<Token>>) -> String {
-    let mut buf = String::new();
-    while let Some(c) = toks.peek() {
-        if !c.kind.is_ascii_digit() {
-            break;
-        }
-        let tok = toks.next().unwrap();
-        buf.push(tok.kind);
-    }
-    buf
-}
